#include "nvfs.h"

void nvfs_free_blocks(struct nvfs_superblock *nvs, unsigned long i, unsigned long n)
{
	unsigned long group, offset;
	uint8_t *bitmap;
	uint8_t *flush_start, *flush_end;
	struct nvfs_group *g;
	struct nvfs_group_info *gi;
	unsigned long nn = n;

next:
	group = i >> log2_group_size(nvs);
	offset = (i & ((1UL << log2_group_size(nvs)) - 1)) >> nvs->log2_page_size;

	if (unlikely(group >= nvs->n_groups)) {
		nvfs_error(nvs, "freeing invalid block %lu", i);
		return;
	}

	bitmap = p_bitmap(nvs, group);
	gi = &nvs->group_info[group];

	spin_lock(&gi->lock);

	flush_start = bitmap + offset / 8;

next_in_group:
	if (!__test_and_clear_bit_le(offset, bitmap)) {
		nvfs_error(nvs, "block bit %lu already cleared", i);
		spin_unlock(&gi->lock);
		return;
	}

	g = p_group(nvs, group);

	gi->n_free_blocks++;

	n--;
	i += 1UL << nvs->log2_page_size;
	if (n && likely(offset != (1UL << log2_group_size(nvs)) - 1)) {
		/* !!! TODO: spin_is_contended is dummy on most architectuers */
		if (unlikely(spin_is_contended(&gi->lock)) && unlikely(!(n & 0xf)))
			goto drop_spinlock;
		offset++;
		goto next_in_group;
	}

drop_spinlock:
	flush_end = bitmap + offset / 8;
	arch_wb_cache_pmem(flush_start, flush_end - flush_start + 1);

	nv_store(g->n_free_blocks, cpu_to_le32(gi->n_free_blocks));

	spin_unlock(&gi->lock);

	if (unlikely(n != 0))
		goto next;

	percpu_counter_add(&nvs->free_blocks_cnt, nn);
}

unsigned long nvfs_free_run(struct nvfs_superblock *nvs, __le64 *run, unsigned long n, bool zero)
{
	/*
	 * TODO: optimization - take the lock just one when freeing
	 * in the same group
	 */
	unsigned long freed = 0;
	while (n--) {
		unsigned long to_free = 1;
		unsigned long block_to_free = le64_to_cpu(*run);
		if (block_to_free) {
			while (to_free <= n && le64_to_cpu(run[to_free]) == block_to_free + (to_free << nvs->log2_page_size) && 0)
				to_free++;
			freed += to_free;
			/*printk("free: %lx, %lx\n", block_to_free, to_free);*/
			nvfs_free_blocks(nvs, block_to_free, to_free);
			if (zero) {
				unsigned long i;
				for (i = 0; i < to_free; i++)
					nv_store(run[i], cpu_to_le64(0));
			}
			n -= to_free - 1;
		}
		run += to_free;
	}
	return freed;
}

static unsigned long block_try_group(struct nvfs_superblock *nvs, unsigned long group, unsigned long offset, unsigned long limit)
{
	struct nvfs_group *g;
	struct nvfs_group_info *gi;
	uint8_t *bitmap;
	unsigned long o;

	if (unlikely(WARN_ON(group >= nvs->n_groups)))
		return ~0;

	g = p_group(nvs, group);
	gi = &nvs->group_info[group];
	bitmap = p_bitmap(nvs, group);

	if (unlikely(READ_ONCE(g->n_free_blocks) < limit))
		return ~0;

search_again:
	o = nvfs_find_bit_in_bmp(bitmap, offset, 1UL << log2_pages_per_group(nvs));
	if (likely(o != ~0)) {
		spin_lock(&gi->lock);

		if (unlikely(__test_and_set_bit_le(o, bitmap))) {
			/* we have lost the race */
			spin_unlock(&gi->lock);
			goto search_again;
		}
		arch_wb_cache_pmem(bitmap + o / 8, 1);

		if (unlikely(o == 1) || (unlikely(!o) && unlikely(!group))) {
			spin_unlock(&gi->lock);
			nvfs_error(nvs, "allocating reserved block %lu", (group << log2_pages_per_group(nvs)) | o);
			return ~0;
		}

		gi->n_free_blocks--;
		nv_store(g->n_free_blocks, cpu_to_le32(gi->n_free_blocks));
		wmb();
		spin_unlock(&gi->lock);
		percpu_counter_dec(&nvs->free_blocks_cnt);
	}

	return o;
}

uint64_t nvfs_new_block(struct nvfs_superblock *nvs)
{
	unsigned long group, offset, o, i;
	struct nvfs_per_cpu *npc = raw_cpu_ptr(nvs->npc);

	i = READ_ONCE(npc->current_block);
	group = i >> log2_pages_per_group(nvs);
	offset = i & ((1UL << log2_pages_per_group(nvs)) - 1);

	o = block_try_group(nvs, group, offset, 1);
	if (unlikely(o == ~0)) {
		unsigned long x;
		uint64_t freei = percpu_counter_read_positive(&nvs->free_blocks_cnt);
		uint64_t average;

		average = div64_u64(freei, nvs->n_groups);

		offset = nvfs_get_random(1UL << log2_pages_per_group(nvs));
		for (x = 0; x < nvs->n_groups; x++) {
			group = nvfs_get_random(nvs->n_groups);
			o = block_try_group(nvs, group, offset, average);
			if (o != ~0)
				goto found;
		}
		for (group = 0; group < nvs->n_groups; group++) {
			o = block_try_group(nvs, group, offset, 1);
			if (o != ~0)
				goto found;
		}
		return 0;
	}

found:
	i = (group << log2_pages_per_group(nvs)) | o;

	if (unlikely(i + 1 < nvs->n_groups << log2_pages_per_group(nvs)))
		WRITE_ONCE(npc->current_block, i + 1);
	else
		WRITE_ONCE(npc->current_block, 0);

	return i << nvs->log2_page_size;
}
